import numpy as np
import pytest
import xarray as xr

from xgcm import Grid
from xgcm.autogenerate import generate_grid_ds

from xmip.grids import (
    _interp_vertex_to_bounds,
    _parse_bounds_vertex,
    combine_staggered_grid,
    create_full_grid,
    detect_shift,
    distance,
    distance_deg,
    recreate_metrics,
)


def _add_small_rand(da):
    return da + (np.random.rand(*da.shape) * 0.05)


def _test_data(grid_label="gn", z_axis=True):
    xt = np.arange(4) + 1
    yt = np.arange(5) + 1
    zt = np.arange(6) + 1

    x = xr.DataArray(xt, coords=[("x", xt)])
    y = xr.DataArray(yt, coords=[("y", yt)])
    lev = xr.DataArray(zt, coords=[("lev", zt)])

    # Need to add a tracer here to get the tracer dimsuffix
    coords = [("x", x.data), ("y", y.data)]
    data = np.random.rand(len(xt), len(yt))
    dims = ["x", "y"]

    if z_axis:
        coords.append(("lev", lev.data))
        data = np.random.rand(len(x), len(y), len(lev))
        dims = ["x", "y", "lev"]

    tr = xr.DataArray(
        data,
        dims=dims,
        coords=coords,
    )

    lon_raw = xr.DataArray(xt, coords=[("x", xt)])
    lat_raw = xr.DataArray(yt, coords=[("y", yt)])
    lon = lon_raw * xr.ones_like(lat_raw)
    lat = xr.ones_like(lon_raw) * lat_raw

    lon_bounds_e = lon + 0.5
    lon_bounds_w = lon - 0.5 + (np.random.rand(*lon.shape) * 0.05)
    lat_bounds_n = lat + 0.5 + (np.random.rand(*lon.shape) * 0.05)
    lat_bounds_s = lat - 0.5 + (np.random.rand(*lon.shape) * 0.05)

    lon_bounds = xr.concat(
        [_add_small_rand(lon_bounds_w), _add_small_rand(lon_bounds_w)], dim="bnds"
    )
    lat_bounds = xr.concat(
        [_add_small_rand(lat_bounds_s), _add_small_rand(lat_bounds_n)], dim="bnds"
    )

    if z_axis:
        lev_bounds = xr.concat(
            [_add_small_rand(lev - 0.5), _add_small_rand(lev + 0.5)], dim="bnds"
        )

    lon_verticies = xr.concat(
        [
            _add_small_rand(lon_bounds_e),
            _add_small_rand(lon_bounds_e),
            _add_small_rand(lon_bounds_w),
            _add_small_rand(lon_bounds_w),
        ],
        dim="vertex",
    )
    lat_verticies = xr.concat(
        [
            _add_small_rand(lat_bounds_s),
            _add_small_rand(lat_bounds_n),
            _add_small_rand(lat_bounds_n),
            _add_small_rand(lat_bounds_s),
        ],
        dim="vertex",
    )

    ds = xr.Dataset({"base": tr})

    dataset_coords = dict(
        lon=lon,
        lat=lat,
        lon_bounds=lon_bounds,
        lat_bounds=lat_bounds,
        lon_verticies=lon_verticies,
        lat_verticies=lat_verticies,
    )

    if z_axis:
        dataset_coords["lev_bounds"] = lev_bounds

    ds = ds.assign_coords(dataset_coords)
    ds.attrs["source_id"] = "test_model"
    ds.attrs["grid_label"] = grid_label
    ds.attrs["variable_id"] = "base"
    return ds


def test_parse_bounds_vertex():
    lon_b = xr.DataArray(np.array([0, 1, 2, 3]), dims=["vertex"])
    lat_b = xr.DataArray(np.array([10, 11, 12, 13]), dims=["vertex"])

    data = np.random.rand(4)

    da = xr.DataArray(
        data, dims=["vertex"], coords={"lon_verticies": lon_b, "lat_verticies": lat_b}
    )
    test = _parse_bounds_vertex(da, "vertex", position=[0, 3])
    print(test)
    expected = (da.isel(vertex=0).load().data, da.isel(vertex=3).load().data)
    print(expected)
    assert test == expected


def test_interp_vertex_to_bounds():
    da = xr.DataArray(np.arange(4), dims=["vertex"])
    # test interp on the y axis
    expected = xr.DataArray(np.array([1.5, 1.5]), dims=["bnds"])
    xr.testing.assert_equal(_interp_vertex_to_bounds(da, "y"), expected)
    # test interp on the x axis
    expected = xr.DataArray(np.array([0.5, 2.5]), dims=["bnds"])
    xr.testing.assert_equal(_interp_vertex_to_bounds(da, "x"), expected)


def test_distance_deg():
    lon0, lat0, lon1, lat1 = 120, 30, 121, 31
    delta_lon, delta_lat = distance_deg(lon0, lat0, lon1, lat1)
    assert delta_lon == 1.0
    assert delta_lat == 1.0

    lon0, lat0, lon1, lat1 = 360, 30, 1, 31
    delta_lon, delta_lat = distance_deg(lon0, lat0, lon1, lat1)
    assert delta_lon == 1.0
    assert delta_lat == 1.0

    lon0, lat0, lon1, lat1 = 300, 30, 301, 30.09
    delta_lon, delta_lat = distance_deg(lon0, lat0, lon1, lat1)
    assert delta_lon == 1.0
    assert delta_lat == 0.0


@pytest.mark.parametrize("lon", [0, 90, 120])
@pytest.mark.parametrize("lat", [0, 10, 45])
def test_distance(lon, lat):
    Re = 6.378e6
    # test straight lat line
    lon0, lat0, lon1, lat1 = lon, lat, lon, lat + 1
    dist = distance(lon0, lat0, lon1, lat1)
    np.testing.assert_allclose(dist, Re * (np.pi * 1.0 / 180))

    # test straight lon line
    lon0, lat0, lon1, lat1 = lon, lat, lon + 1, lat
    dist = distance(lon0, lat0, lon1, lat1)
    np.testing.assert_allclose(
        dist, Re * (np.pi * 1.0 / 180) * np.cos(np.pi * lat0 / 180)
    )


# TODO: inner and outer (needs to be implemented in xgcm autogenerate first)
@pytest.mark.parametrize("xshift", ["left", "right"])
@pytest.mark.parametrize("yshift", ["left", "right"])
@pytest.mark.parametrize("z_axis", [True, False])
def test_recreate_metrics(xshift, yshift, z_axis):

    # reconstruct all the metrics by hand and compare to inferred output

    # * For now this is a regular lon lat grid. Might need to add some tests for more complex grids.
    # Then again. This will not do a great job for those....

    # create test dataset
    ds = _test_data(z_axis=z_axis)

    # TODO: generalize so this also works with e.g. zonal average sections (which dont have a X axis)
    coord_dict = {"X": "x", "Y": "y"}
    if z_axis:
        coord_dict["Z"] = "lev"

    ds_full = generate_grid_ds(
        ds,
        coord_dict,
        position={"X": ("center", xshift), "Y": ("center", yshift)},
    )

    grid = Grid(ds_full)

    ds_metrics, metrics_dict = recreate_metrics(ds_full, grid)

    if z_axis:
        # Check that the bound values are intact (previously those got alterd due to unexpected behaviour of .assign_coords())
        assert "bnds" in ds_metrics.lev_bounds.dims

    # compute the more complex metrics (I could wrap this into a function I guess?)
    lon0, lon1 = grid.axes["X"]._get_neighbor_data_pairs(ds.lon.load(), xshift)
    lat0, lat1 = grid.axes["X"]._get_neighbor_data_pairs(ds.lat.load(), xshift)
    dx_gx_expected = distance(lon0, lat0, lon1, lat1)

    lon0, lon1 = grid.axes["Y"]._get_neighbor_data_pairs(ds.lon.load(), yshift)
    lat0, lat1 = grid.axes["Y"]._get_neighbor_data_pairs(ds.lat.load(), yshift)
    dy_gy_expected = distance(lon0, lat0, lon1, lat1)

    # corner metrics
    # dx
    if yshift == "left":
        # dx
        lon0, lon1 = grid.axes["X"]._get_neighbor_data_pairs(
            _interp_vertex_to_bounds(ds_metrics.lon_verticies, "y").isel(bnds=0),
            xshift,
        )
        lat0, lat1 = grid.axes["X"]._get_neighbor_data_pairs(
            ds_metrics.lat_bounds.isel(bnds=0), xshift
        )
    elif yshift == "right":
        lon0, lon1 = grid.axes["X"]._get_neighbor_data_pairs(
            _interp_vertex_to_bounds(ds_metrics.lon_verticies, "y").isel(bnds=1),
            xshift,
        )
        lat0, lat1 = grid.axes["X"]._get_neighbor_data_pairs(
            ds_metrics.lat_bounds.isel(bnds=1), xshift
        )
    dx_gxgy_expected = distance(lon0, lat0, lon1, lat1)

    # dy
    if xshift == "left":
        # dx
        lat0, lat1 = grid.axes["Y"]._get_neighbor_data_pairs(
            _interp_vertex_to_bounds(ds_metrics.lat_verticies, "x").isel(bnds=0),
            yshift,
        )
        lon0, lon1 = grid.axes["Y"]._get_neighbor_data_pairs(
            ds_metrics.lon_bounds.isel(bnds=0), yshift
        )
    elif xshift == "right":
        lat0, lat1 = grid.axes["Y"]._get_neighbor_data_pairs(
            _interp_vertex_to_bounds(ds_metrics.lat_verticies, "x").isel(bnds=1),
            yshift,
        )
        lon0, lon1 = grid.axes["Y"]._get_neighbor_data_pairs(
            ds_metrics.lon_bounds.isel(bnds=1), yshift
        )
    dy_gxgy_expected = distance(lon0, lat0, lon1, lat1)

    if xshift == "left":
        vertex_points = [0, 1]
    else:
        vertex_points = [2, 3]
    lon0, lon1 = (
        ds_metrics.lon_verticies.isel(vertex=vertex_points[0]),
        ds_metrics.lon_verticies.isel(vertex=vertex_points[1]),
    )
    lat0, lat1 = (
        ds_metrics.lat_verticies.isel(vertex=vertex_points[0]),
        ds_metrics.lat_verticies.isel(vertex=vertex_points[1]),
    )
    dy_gx_expected = distance(lon0, lat0, lon1, lat1)

    if yshift == "left":
        vertex_points = [0, 3]
    else:
        vertex_points = [1, 2]
    lon0, lon1 = (
        ds_metrics.lon_verticies.isel(vertex=vertex_points[0]),
        ds_metrics.lon_verticies.isel(vertex=vertex_points[1]),
    )
    lat0, lat1 = (
        ds_metrics.lat_verticies.isel(vertex=vertex_points[0]),
        ds_metrics.lat_verticies.isel(vertex=vertex_points[1]),
    )
    dx_gy_expected = distance(lon0, lat0, lon1, lat1)

    if z_axis:
        dz_t_expected = ds.lev_bounds.diff("bnds").squeeze().data
    else:
        dz_t_expected = None

    for var, expected in [
        ("dz_t", dz_t_expected),
        (
            "dx_t",
            distance(
                ds_metrics.lon_bounds.isel(bnds=0).data,
                ds_metrics.lat.data,
                ds_metrics.lon_bounds.isel(bnds=1).data,
                ds_metrics.lat.data,
            ),
        ),
        (
            "dy_t",
            distance(
                ds_metrics.lon.data,
                ds_metrics.lat_bounds.isel(bnds=0).data,
                ds_metrics.lon.data,
                ds_metrics.lat_bounds.isel(bnds=1).data,
            ),
        ),
        ("dx_gx", dx_gx_expected),
        ("dy_gy", dy_gy_expected),
        ("dy_gx", dy_gx_expected),
        ("dx_gy", dx_gy_expected),
        ("dy_gxgy", dy_gxgy_expected),
        ("dx_gxgy", dx_gxgy_expected),
    ]:
        if expected is not None:
            print(var)
            control = ds_metrics[var].data
            if expected.shape != control.shape:
                control = control.T
            np.testing.assert_allclose(control, expected)

    if z_axis:
        assert set(["X", "Y", "Z"]).issubset(set(metrics_dict.keys()))
    else:
        assert set(["X", "Y"]).issubset(set(metrics_dict.keys()))
        assert "Z" not in list(metrics_dict.keys())


# TODO: inner and outer (needs to be implemented in xgcm autogenerate first)
@pytest.mark.parametrize("xshift", ["left", "center", "right"])
@pytest.mark.parametrize("yshift", ["left", "center", "right"])
def test_detect_shift(xshift, yshift):

    # create base dataset (tracer)
    ds_base = _test_data()

    # create the maybe shifted dataset
    ds = ds_base.copy()
    if xshift == "left":
        ds["lon"] = ds["lon"] - 0.5
    elif xshift == "right":
        ds["lon"] = ds["lon"] + 0.5

    if yshift == "left":
        ds["lat"] = ds["lat"] - 0.5
    elif yshift == "right":
        ds["lat"] = ds["lat"] + 0.5
    assert detect_shift(ds_base, ds, "X") == xshift
    assert detect_shift(ds_base, ds, "Y") == yshift

    # repeat with very small shifts (these should not be detected)
    ds = ds_base.copy()
    if xshift == "left":
        ds["lon"] = ds["lon"] - 0.05
    elif xshift == "right":
        ds["lon"] = ds["lon"] + 0.05

    if yshift == "left":
        ds["lat"] = ds["lat"] - 0.05
    elif yshift == "right":
        ds["lat"] = ds["lat"] + 0.05
    assert detect_shift(ds_base, ds, "X") == "center"
    assert detect_shift(ds_base, ds, "Y") == "center"


@pytest.mark.parametrize("xshift", ["left", "right"])
@pytest.mark.parametrize("yshift", ["left", "right"])
@pytest.mark.parametrize("grid_label", ["gr", "gn"])
def test_create_full_grid(xshift, yshift, grid_label):
    ds_base = _test_data(grid_label=grid_label)
    grid_dict = {"test_model": {grid_label: {"axis_shift": {"X": xshift, "Y": yshift}}}}
    # TODO: This should be specific to the grid_label: e.g grid_dict = {'model':{'gr':{'axis_shift':{'X':'left}}}}

    ds_full = create_full_grid(ds_base, grid_dict=grid_dict)

    shift_dict = {"left": -0.5, "right": 0.5}

    assert ds_full["x"].attrs["axis"] == "X"
    assert ds_full["x_" + xshift].attrs["axis"] == "X"
    assert ds_full["x_" + xshift].attrs["c_grid_axis_shift"] == shift_dict[xshift]
    assert ds_full["y"].attrs["axis"] == "Y"
    assert ds_full["y_" + yshift].attrs["axis"] == "Y"
    assert ds_full["y_" + yshift].attrs["c_grid_axis_shift"] == shift_dict[yshift]
    # TODO: integrate the vertical
    # assert ds_full["lev"].attrs["axis"] == "Z"

    # I might want to loosen this later and switch to a uniform naming
    # E.g. use x_g for the x dimension on the x gridface, no matter if its left or right...
    # TODO: Check upstream in xgcm
    # Once that is done I
    assert "x_" + xshift in ds_full.dims
    assert "y_" + yshift in ds_full.dims

    # test error handling
    with pytest.warns(UserWarning):
        ds_none = create_full_grid(
            ds_base, grid_dict=None
        )  # the synthetic dataset is not in the default dict.
    assert ds_none is None


@pytest.mark.parametrize("recalculate_metrics", [True, False])
@pytest.mark.parametrize("xshift", ["left", "right"])
@pytest.mark.parametrize("yshift", ["left", "right"])
@pytest.mark.parametrize("grid_label", ["gr", "gn"])
def test_combine_staggered_grid(recalculate_metrics, xshift, yshift, grid_label):
    ds_base = _test_data(grid_label=grid_label)

    # create the maybe shifted dataset
    ds = ds_base.copy()
    ds = ds.rename({"base": "other"})
    ds.attrs["variable_id"] = "other"
    if xshift == "left":
        ds["lon"] = ds["lon"] - 0.5
    elif xshift == "right":
        ds["lon"] = ds["lon"] + 0.5

    if yshift == "left":
        ds["lat"] = ds["lat"] - 0.5
    elif yshift == "right":
        ds["lat"] = ds["lat"] + 0.5
    grid_dict = {"test_model": {grid_label: {"axis_shift": {"X": xshift, "Y": yshift}}}}

    for other_ds in [ds, [ds]]:
        grid, ds_combined = combine_staggered_grid(
            ds_base,
            other_ds,
            grid_dict=grid_dict,
            recalculate_metrics=recalculate_metrics,
        )

        for axis, shift in zip(["X", "Y"], [xshift, yshift]):
            # make sure the correct dim is in the added dataset
            assert grid.axes[axis].coords[shift] in ds_combined["other"].dims
            # and also that none of the other are in there
            assert all(
                [
                    di not in ds_combined["other"].dims
                    for dd, di in grid.axes[axis].coords.items()
                    if dd != shift
                ]
            )
        # check if metrics are correctly parsed
        if recalculate_metrics:
            for axis in ["X", "Y"]:
                for metric in ["_t", "_gx", "_gy", "_gxgy"]:
                    assert f"d{axis.lower()}{metric}" in list(ds_combined.coords)

    # Test error handling
    with pytest.warns(UserWarning):
        grid_none, ds_combined_none = combine_staggered_grid(
            ds_base,
            ds,
            grid_dict=None,
            recalculate_metrics=recalculate_metrics,
        )
    assert ds_combined_none is None
    assert grid_none is None



# This module tests data directly from the pangeo google cloud storage.
# Tests are meant to be more high level and also serve to document known problems (see skip statements).

import fsspec
import numpy as np
import pytest
import xarray as xr

from xmip.grids import combine_staggered_grid
from xmip.preprocessing import _desired_units, _drop_coords, combined_preprocessing
from xmip.utils import google_cmip_col, model_id_match


pytest.importorskip("gcsfs")


def diagnose_duplicates(data):
    """displays non-unique entries in data"""
    _, idx = np.unique(data, return_index=True)
    missing = np.array([i for i in np.arange(len(data)) if i not in idx])
    if len(missing) > 0:
        missing_values = data[missing]
        raise ValueError(f"Duplicate Values ({missing_values}) found")


def data(
    source_id, variable_id, experiment_id, grid_label, use_intake_esm, catalog="main"
):
    zarr_kwargs = {
        "consolidated": True,
        # "decode_times": False,
        "decode_times": True,
        "use_cftime": True,
    }

    cat = google_cmip_col(catalog=catalog).search(
        source_id=source_id,
        experiment_id=experiment_id,
        variable_id=variable_id,
        # member_id="r1i1p1f1",
        table_id="Omon",
        grid_label=grid_label,
    )

    if len(cat.df["zstore"]) > 0:
        if use_intake_esm:
            ddict = cat.to_dataset_dict(
                zarr_kwargs=zarr_kwargs,
                preprocess=combined_preprocessing,
                storage_options={"token": "anon"},
            )
            _, ds = ddict.popitem()
        else:
            # debugging options
            # @charlesbluca suggested this to make this work in GHA
            # https://github.com/jbusecke/xmip/pull/62#issuecomment-741928365
            mm = fsspec.get_mapper(
                cat.df["zstore"][0]
            )  # think you can pass in storage options here as well?
            ds_raw = xr.open_zarr(mm, **zarr_kwargs)
            ds = combined_preprocessing(ds_raw)
    else:
        ds = None

    return ds, cat


def all_models():
    df = google_cmip_col().df
    all_models = df["source_id"].unique()
    all_models = tuple(np.sort(all_models))
    # all_models = tuple(["EC-Earth3"])
    return all_models


# test_models = ["CESM2-FV2", "GFDL-CM4"]
test_models = all_models()


def pytest_generate_tests(metafunc):
    # This is called for every test. Only get/set command line arguments
    # if the argument is specified in the list of test "fixturenames".

    for name in ["vi", "gl", "ei", "cat"]:

        option_value = getattr(metafunc.config.option, name)

        if isinstance(option_value, str):
            option_value = [option_value]

        if name in metafunc.fixturenames and option_value is not None:
            metafunc.parametrize(name, option_value)


# print(f"\n\n\n\n$$$$$$$ All available models: {all_models()}$$$$$$$\n\n\n\n")

# Combine the input parameters according to command line input

# --- Most basic test --- #

# Try to combine some of the failures

# We dont support these at all
not_supported_failures = [
    ("AWI-ESM-1-1-LR", "*", "*", "gn"),
    ("AWI-CM-1-1-MR", "*", "*", "gn"),
]

# basic problems when trying to concat with intake-esm
intake_concat_failures = [
    (
        "CanESM5",
        [
            "uo",
            "so",
            "thetao",
        ],
        "ssp245",
        "gn",
    ),
    (
        "CanESM5",
        ["zos"],
        [
            "ssp245",
            "ssp585",
        ],
        "gn",
    ),
    (
        "E3SM-1-0",
        ["so", "o2", "zos"],
        ["historical", "ssp585"],
        "gr",
    ),  # issues with time concatenation
    (
        "IPSL-CM6A-LR",
        ["thetao", "o2", "so"],
        "historical",
        "gn",
    ),  # IPSL has an issue with `lev` dims concatting]
    (
        "NorESM2-MM",
        ["uo", "so"],
        "historical",
        "gr",
    ),  # time concatting
    (
        "NorESM2-MM",
        ["so"],
        "historical",
        "gn",
    ),
]


# this fixture has to be redifined every time to account for different fail cases for each test
@pytest.fixture
def spec_check_dim_coord_values_wo_intake(request, gl, vi, ei, cat):
    expected_failures = not_supported_failures + [
        ("FGOALS-f3-L", ["thetao"], "piControl", "gn"),
        # (
        #     "GFDL-CM4",
        #     "thetao",
        #     "historical",
        #     "gn",
        # ),  # this should not fail and should trigger an xpass (I just use this for dev purposes to check
        #     # the strict option)
    ]
    spec = (request.param, vi, ei, gl, cat)
    request.param = spec
    if model_id_match(expected_failures, request.param[0:-1]):
        request.node.add_marker(pytest.mark.xfail(strict=True))
    return request


@pytest.mark.parametrize(
    "spec_check_dim_coord_values_wo_intake", test_models, indirect=True
)
def test_check_dim_coord_values_wo_intake(
    spec_check_dim_coord_values_wo_intake,
):
    (
        source_id,
        variable_id,
        experiment_id,
        grid_label,
        catalog,
    ) = spec_check_dim_coord_values_wo_intake.param

    # there must be a better way to build this at the class level and then tear it down again
    # I can probably get this done with fixtures, but I dont know how atm
    ds, _ = data(
        source_id, variable_id, experiment_id, grid_label, False, catalog=catalog
    )

    if ds is None:
        pytest.skip(
            f"No data found for {source_id}|{variable_id}|{experiment_id}|{grid_label}"
        )

    # Check for dim duplicates
    # check all dims for duplicates
    # for di in ds.dims:
    # for now only test a subset of the dims. TODO: Add the bounds once they
    # are cleaned up.
    for di in ["x", "y", "lev", "time"]:
        if di in ds.dims:
            diagnose_duplicates(ds[di].load().data)
            assert len(ds[di]) == len(np.unique(ds[di]))
            if di != "time":  # these tests do not make sense for decoded time
                assert np.all(~np.isnan(ds[di]))
                assert np.all(ds[di].diff(di) >= 0)

    assert ds.lon.min().load() >= 0
    assert ds.lon.max().load() <= 360
    if "lon_bounds" in ds.variables:
        assert ds.lon_bounds.min().load() >= 0
        assert ds.lon_bounds.max().load() <= 361
    assert ds.lat.min().load() >= -90
    assert ds.lat.max().load() <= 90
    # make sure lon and lat are 2d
    assert len(ds.lon.shape) == 2
    assert len(ds.lat.shape) == 2
    for co in _drop_coords:
        if co in ds.dims:
            assert co not in ds.coords

    # Check unit conversion
    for var, expected_unit in _desired_units.items():
        if var in ds.variables:
            unit = ds[var].attrs.get("units")
            if unit:
                assert unit == expected_unit


# this fixture has to be redifined every time to account for different fail cases for each test
@pytest.fixture
def spec_check_dim_coord_values(request, gl, vi, ei, cat):
    expected_failures = (
        not_supported_failures
        + intake_concat_failures
        + [
            ("NorESM2-MM", ["uo", "zos"], "historical", "gn"),
            ("NorESM2-MM", "thetao", "historical", "gn"),
            ("NorESM2-MM", "thetao", "historical", "gr"),
            ("FGOALS-f3-L", ["thetao"], "piControl", "gn"),
        ]
    )
    spec = (request.param, vi, ei, gl, cat)
    request.param = spec
    if model_id_match(expected_failures, request.param[0:-1]):
        request.node.add_marker(pytest.mark.xfail(strict=True))
    return request


@pytest.mark.parametrize("spec_check_dim_coord_values", test_models, indirect=True)
def test_check_dim_coord_values(
    spec_check_dim_coord_values,
):
    (
        source_id,
        variable_id,
        experiment_id,
        grid_label,
        catalog,
    ) = spec_check_dim_coord_values.param
    # there must be a better way to build this at the class level and then tear it down again
    # I can probably get this done with fixtures, but I dont know how atm
    ds, cat = data(
        source_id, variable_id, experiment_id, grid_label, True, catalog=catalog
    )

    if ds is None:
        pytest.skip(
            f"No data found for {source_id}|{variable_id}|{experiment_id}|{grid_label}"
        )

    # Check for dim duplicates
    # check all dims for duplicates
    # for di in ds.dims:
    # for now only test a subset of the dims. TODO: Add the bounds once they
    # are cleaned up.
    for di in ["x", "y", "lev", "time"]:
        if di in ds.dims:
            diagnose_duplicates(ds[di].load().data)
            assert len(ds[di]) == len(np.unique(ds[di]))
            if di != "time":  # these tests do not make sense for decoded time
                assert np.all(~np.isnan(ds[di]))
                assert np.all(ds[di].diff(di) >= 0)

    assert ds.lon.min().load() >= 0
    assert ds.lon.max().load() <= 360
    if "lon_bounds" in ds.variables:
        assert ds.lon_bounds.min().load() >= 0
        assert ds.lon_bounds.max().load() <= 361
    assert ds.lat.min().load() >= -90
    assert ds.lat.max().load() <= 90
    # make sure lon and lat are 2d
    assert len(ds.lon.shape) == 2
    assert len(ds.lat.shape) == 2
    for co in _drop_coords:
        if co in ds.dims:
            assert co not in ds.coords


# --- Specific Bound Coords Test -----


# this fixture has to be redifined every time to account for different fail cases for each test
@pytest.fixture
def spec_check_bounds_verticies(request, gl, vi, ei, cat):
    expected_failures = (
        not_supported_failures
        + intake_concat_failures
        + [
            ("FGOALS-f3-L", ["thetao", "so", "uo", "zos"], "*", "gn"),
            ("FGOALS-g3", ["thetao", "so", "uo", "zos"], "*", "gn"),
            ("NorESM2-MM", ["thetao", "uo", "zos"], "historical", "gn"),
            ("NorESM2-MM", ["thetao", "so"], "historical", "gr"),
            ("IPSL-CM6A-LR", ["thetao", "o2"], "historical", "gn"),
            ("IITM-ESM", ["so", "uo", "thetao"], "piControl", "gn"),
            ("GFDL-CM4", "uo", "*", "gn"),
        ]
    )
    spec = (request.param, vi, ei, gl, cat)
    request.param = spec
    if model_id_match(expected_failures, request.param[0:-1]):
        request.node.add_marker(pytest.mark.xfail(strict=True))
    return request


@pytest.mark.parametrize("spec_check_bounds_verticies", test_models, indirect=True)
def test_check_bounds_verticies(spec_check_bounds_verticies):
    (
        source_id,
        variable_id,
        experiment_id,
        grid_label,
        catalog,
    ) = spec_check_bounds_verticies.param
    ds, cat = data(
        source_id, variable_id, experiment_id, grid_label, True, catalog=catalog
    )

    if ds is None:
        pytest.skip(
            f"No data found for {source_id}|{variable_id}|{experiment_id}|{grid_label}"
        )

    if "vertex" in ds.dims:
        np.testing.assert_allclose(ds.vertex.data, np.arange(4))

    # Check for existing bounds and verticies
    for co in ["lon_bounds", "lat_bounds", "lon_verticies", "lat_verticies"]:
        assert co in ds.coords
        # make sure that all other dims are eliminated from the bounds.
        assert (set(ds[co].dims) - set(["bnds", "vertex"])) == set(["x", "y"])

    # Check the order of the vertex
    # Ill only check these south of the Arctic for now. Up there
    # things are still weird.
    test_ds = ds.where(abs(ds.lat) <= 40, drop=True)

    vertex_lon_diff1 = test_ds.lon_verticies.isel(
        vertex=3
    ) - test_ds.lon_verticies.isel(vertex=0)
    vertex_lon_diff2 = test_ds.lon_verticies.isel(
        vertex=2
    ) - test_ds.lon_verticies.isel(vertex=1)
    vertex_lat_diff1 = test_ds.lat_verticies.isel(
        vertex=1
    ) - test_ds.lat_verticies.isel(vertex=0)
    vertex_lat_diff2 = test_ds.lat_verticies.isel(
        vertex=2
    ) - test_ds.lat_verticies.isel(vertex=3)
    for vertex_diff in [vertex_lon_diff1, vertex_lon_diff2]:
        assert (vertex_diff <= 0).sum() <= (3 * len(vertex_diff.y))
        # allowing for a few rows to be negative

    for vertex_diff in [vertex_lat_diff1, vertex_lat_diff2]:
        assert (vertex_diff <= 0).sum() <= (5 * len(vertex_diff.x))
        # allowing for a few rows to be negative
    # This is just to make sure that not the majority of values is negative or zero.

    # Same for the bounds:
    lon_diffs = test_ds.lon_bounds.diff("bnds")
    lat_diffs = test_ds.lat_bounds.diff("bnds")

    assert (lon_diffs <= 0).sum() <= (5 * len(lon_diffs.y))
    assert (lat_diffs <= 0).sum() <= (5 * len(lat_diffs.y))


# --- xgcm grid specific tests --- #
# this fixture has to be redifined every time to account for different fail cases for each test
@pytest.fixture
def spec_check_grid(request, gl, vi, ei, cat):
    expected_failures = (
        not_supported_failures
        + intake_concat_failures
        + [
            ("CMCC-ESM2", "*", "*", "gn"),
            ("CMCC-CM2-SR5", "*", "*", "gn"),
            ("CMCC-CM2-HR4", "*", "*", "gn"),
            ("FGOALS-f3-L", "*", "*", "gn"),
            ("FGOALS-g3", "*", "*", "gn"),
            ("E3SM-1-0", ["so", "thetao", "o2"], "*", "gn"),
            (
                "E3SM-1-0",
                ["zos"],
                ["historical", "ssp585", "ssp245", "ssp370", "esm-hist"],
                "gr",
            ),
            (
                "EC-Earth3-AerChem",
                ["so", "thetao", "zos"],
                ["historical", "piControl", "ssp370"],
                "gn",
            ),
            ("EC-Earth3-Veg", "*", "historical", "gr"),
            ("EC-Earth3-CC", "*", "*", "gn"),
            ("MPI-ESM-1-2-HAM", "*", "*", "gn"),
            ("NorESM2-MM", "*", "historical", "gn"),
            ("NorESM2-MM", ["thetao", "so", "uo"], "historical", "gr"),
            ("IITM-ESM", "*", "*", "gn"),
            ("GFDL-CM4", ["uo"], "*", "gn"),
            ("IPSL-CM5A2-INCA", "*", "*", "gn"),
            ("IPSL-CM6A-LR-INCA", "*", "*", "gn"),
        ]
    )
    spec = (request.param, vi, ei, gl, cat)
    request.param = spec
    if model_id_match(expected_failures, request.param[0:-1]):
        request.node.add_marker(pytest.mark.xfail(strict=True, reason=""))
    return request


@pytest.mark.parametrize("spec_check_grid", test_models, indirect=True)
def test_check_grid(
    spec_check_grid,
):
    source_id, variable_id, experiment_id, grid_label, catalog = spec_check_grid.param

    ds, cat = data(
        source_id, variable_id, experiment_id, grid_label, True, catalog=catalog
    )

    if ds is None:
        pytest.skip(
            f"No data found for {source_id}|{variable_id}|{experiment_id}|{grid_label}"
        )

    # This is just a rudimentary test to see if the creation works
    staggered_grid, ds_staggered = combine_staggered_grid(ds, recalculate_metrics=True)

    assert ds_staggered is not None
    #
    if "lev" in ds_staggered.dims:
        assert "bnds" in ds_staggered.lev_bounds.dims

    for axis in ["X", "Y"]:
        for metric in ["_t", "_gx", "_gy", "_gxgy"]:
            assert f"d{axis.lower()}{metric}" in list(ds_staggered.coords)
    # TODO: Include actual test to combine variables



import random
import string

import numpy as np
import pytest
import xarray as xr
import xesmf

from xmip.postprocessing import (
    _construct_and_promote_member_id,
    _parse_metric,
    combine_datasets,
    concat_experiments,
    concat_members,
    interpolate_grid_label,
    match_metrics,
    merge_variables,
    pick_first_member,
)


@pytest.fixture(params=["metric", "other"])
def metricname(request):
    return request.param


def random_ds(time_coords=False, attrs={}):
    """Create random dataset"""
    nx, ny, nz, nt = (3, 2, 4, 6)
    data = np.random.rand(nx, ny, nz, nt)
    da = xr.DataArray(data, dims=["x", "y", "z", "time"])
    if time_coords:
        da = da.assign_coords(time=xr.cftime_range("2000", periods=nt))
    ds = da.to_dataset(name="data")
    ds.attrs = attrs
    return ds


def test_parse_metric(metricname):
    # create a dataset
    ds = random_ds()
    # create a metric dataset
    ds_metric = random_ds().isel(z=0, time=0).rename({"data": metricname})
    metric = ds_metric[metricname]
    metric.attrs.update({"check": "carry"})

    # parse the metric
    ds_parsed = _parse_metric(ds, metric)
    assert metricname in ds_parsed.coords
    xr.testing.assert_allclose(ds_parsed[metricname].reset_coords(drop=True), metric)
    assert (
        ds_parsed[metricname].attrs["parsed_with"]
        == "xmip/postprocessing/_parse_metric"
    )
    # check that existing attrs are conserved
    assert ds_parsed[metricname].attrs["check"] == "carry"


def test_parse_metric_exceptions(metricname):
    # create a dataset
    ds = random_ds()
    # create a metric dataset
    ds_metric = random_ds().isel(z=0, time=0).rename({"data": metricname})

    # provide dataset instead of dataarray
    with pytest.raises(ValueError):
        _parse_metric(ds, ds_metric)


def test_parse_metric_exceptions_input_name():
    # create a dataset
    ds = random_ds()
    # create a metric dataset
    ds_metric = random_ds().isel(z=0, time=0)
    # set attributes
    ds.attrs = {"activity_id": "a"}

    # provide dataarray without name
    with pytest.warns(RuntimeWarning) as warninfo:
        da_metric_nameless = ds_metric["data"]
        da_metric_nameless.name = None

        _parse_metric(ds, da_metric_nameless)
    assert (
        warninfo[0].message.args[0]
        == "a.none.none.none.none.none.none.none.none:`metric` has no name. This might lead to problems down the line."
    )


def test_parse_metric_exception_dim_length():
    metricname = "metric"
    # create a dataset
    ds = random_ds()
    # create a metric dataset
    ds_metric = random_ds().isel(z=0, time=0).rename({"data": metricname})
    # set attributes
    ds.attrs = {"activity_id": "a", "grid_label": "g"}

    # provide dataarray with non-matching dimensions
    with pytest.raises(ValueError) as execinfo:
        _parse_metric(ds, ds_metric.isel(x=slice(0, -1), y=slice(1, None))[metricname])
    msg = "a.none.none.none.none.none.g.none.none:`metric` dimensions ['x:2', 'y:1'] do not match `ds` ['x:3', 'y:2']."
    assert execinfo.value.args[0] == msg


def test_parse_metric_dim_alignment():
    metricname = "metric"
    # create a dataset
    ds = random_ds(time_coords=True)
    # create a metric dataset
    ds_metric = (
        random_ds(time_coords=True)
        .isel(z=0, time=slice(0, -1))
        .rename({"data": metricname})
    )
    # set attributes
    ds.attrs = {"activity_id": "a", "grid_label": "g"}

    print(ds)

    # Allow alignment
    with pytest.warns(UserWarning) as warninfo:
        ds_parsed = _parse_metric(
            ds, ds_metric[metricname], dim_length_conflict="align"
        )

    xr.testing.assert_allclose(ds_parsed.time, ds_metric.time)

    msg = "a.none.none.none.none.none.g.none.none:`metric` dimensions ['time:5'] do not match `ds` ['time:6']. Aligning the data on `inner`"
    assert warninfo[0].message.args[0] == msg


def test_match_metrics(metricname):
    # create a few different datasets
    ds_a = random_ds()
    ds_b = random_ds()
    ds_c = random_ds()
    ds_d = random_ds()
    ds_e = random_ds()

    # Give them cmip attrs
    ds_a.attrs = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }
    ds_b.attrs = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "b",
        "variant_label": "b",
        "version": "b",
    }
    ds_c.attrs = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "b",
        "table_id": "b",
        "variant_label": "a",
        "version": "b",
    }
    ds_d.attrs = {
        "source_id": "a",
        "grid_label": "b",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }
    ds_e.attrs = {
        "source_id": "b",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }

    # now create a metric (which does not vary in time) which matches ds_a
    ds_metric = random_ds().isel(time=0).rename({"data": metricname})
    ds_metric.attrs = ds_a.attrs

    # FIXME: This needs to be factored out
    def _assert_parsed_ds_dict(ddict_parsed, expected, match_keys, strict=True):
        expected = expected.copy()
        for i in match_keys:
            ds_parsed = ddict_parsed[i]
            assert metricname in list(ds_parsed.variables)
            xr.testing.assert_allclose(
                ds_parsed[metricname].reset_coords(drop=True), expected
            )
        if strict:
            for i in [ii for ii in ddict_parsed.keys() if ii not in match_keys]:
                ds_parsed = ddict_parsed[i]
                assert metricname not in ds_parsed.variables

    # With the default options I expect that this gets parsed into a,b,c (all the same source_id and grid_label)
    # but not d and e
    ds_dict = {"a": ds_a, "b": ds_b, "c": ds_c, "d": ds_d, "e": ds_e}
    metric_dict = {"something": ds_metric}
    expected = ds_metric[metricname]

    ds_dict_parsed = match_metrics(ds_dict, metric_dict, [metricname])
    _assert_parsed_ds_dict(ds_dict_parsed, ds_metric[metricname], ["a", "b", "c"])

    # Now change the matching parameter
    ds_dict_parsed = match_metrics(
        ds_dict, metric_dict, match_variables=[metricname], match_attrs="exact"
    )
    _assert_parsed_ds_dict(ds_dict_parsed, expected, ["a"])

    ds_dict_parsed = match_metrics(
        ds_dict,
        metric_dict,
        [metricname],
        match_attrs=["source_id", "grid_label", "experiment_id"],
    )
    _assert_parsed_ds_dict(ds_dict_parsed, expected, ["a", "b"])

    ds_dict_parsed = match_metrics(
        ds_dict,
        metric_dict,
        [metricname],
        match_attrs=["source_id", "grid_label", "variant_label"],
    )
    _assert_parsed_ds_dict(ds_dict_parsed, expected, ["a", "c"])

    # Now give the metric the attributes of e and check
    ds_metric.attrs = ds_e.attrs

    ds_dict = {"a": ds_a, "b": ds_b, "c": ds_c, "d": ds_d, "e": ds_e}
    metric_dict = {"something": ds_metric}

    ds_dict_parsed = match_metrics(ds_dict, metric_dict, match_variables=[metricname])
    _assert_parsed_ds_dict(ds_dict_parsed, ds_metric[metricname], ["e"])

    # Check that a metric with time dimension is never parsed anywhere, except an exact match
    ds_metric = random_ds().rename({"data": metricname})
    ds_metric.attrs = ds_a.attrs

    ds_dict = {"a": ds_a, "b": ds_b, "c": ds_c, "d": ds_d, "e": ds_e}
    metric_dict = {"something": ds_metric}

    ds_dict_parsed = match_metrics(ds_dict, metric_dict, match_variables=[metricname])
    _assert_parsed_ds_dict(ds_dict_parsed, ds_metric[metricname], ["a"])


def test_match_metrics_missing_non_match_attr():
    """This test ensures that as long as the provided `match_metrics` are
    given they will be matched. This is relevant if e.g. the variant label
    has been removed due to merging"""
    metricname = "area"
    ds_a = random_ds()
    ds_a.attrs = {
        "source_id": "a",
        "grid_label": "a",
    }
    ds_metric = random_ds().isel(time=0).rename({"data": metricname})
    ds_metric.attrs = ds_a.attrs

    ds_dict = {"a": ds_a}
    metric_dict = {"something": ds_metric}
    expected = ds_metric[metricname]

    ds_dict_parsed = match_metrics(ds_dict, metric_dict, [metricname])

    assert "a" in ds_dict_parsed.keys()
    # TODO this should be factored out into _assert_parsed_ds_dict from the test above
    xr.testing.assert_allclose(
        expected, ds_dict_parsed["a"][metricname].reset_coords(drop=True)
    )


def test_match_metrics_missing_match_attrs():
    """If one of the `match_attrs` is not in the dataset this should error out"""
    metricname = "area"
    ds_a = random_ds()
    ds_a.attrs = {
        "source_id": "a",
    }
    ds_metric = random_ds().isel(time=0).rename({"data": metricname})
    ds_metric.attrs = ds_a.attrs

    ds_dict = {"a": ds_a}
    metric_dict = {"something": ds_metric}
    with pytest.raises(
        ValueError,
        match="Cannot match datasets because at least one of the datasets does not contain all attributes",
    ):
        match_metrics(ds_dict, metric_dict, [metricname])


def test_match_metrics_closer(metricname):
    # Test to see if a metric dataset with more matching attrs is preferred.

    ds_a = random_ds()
    ds_b = random_ds()
    ds_c = random_ds()

    # Give them cmip attrs
    ds_a.attrs = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }
    ds_b.attrs = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "b",
        "variant_label": "b",
        "version": "b",
    }
    ds_c.attrs = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "b",
        "table_id": "b",
        "variant_label": "a",
        "version": "b",
    }

    ds_metric_a = random_ds().isel(time=0).rename({"data": metricname})
    ds_metric_a.attrs = ds_a.attrs
    ds_metric_c = random_ds().isel(time=0).rename({"data": metricname})
    ds_metric_c.attrs = ds_c.attrs

    ds_dict = {"c": ds_c}
    metric_dict = {
        "exact_c": ds_metric_c,
        "exact_a": ds_metric_a,
    }
    ds_dict_parsed = match_metrics(ds_dict, metric_dict, match_variables=[metricname])
    xr.testing.assert_allclose(
        ds_dict_parsed["c"][metricname].reset_coords(drop=True), ds_metric_c[metricname]
    )
    assert ds_dict_parsed["c"][metricname].attrs["original_key"] == "exact_c"


def test_match_metrics_align_dims():
    metricname = "metric"
    attrs = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
        "variable_id": "a",
    }
    ds = random_ds(time_coords=True)
    ds.attrs = attrs
    ds_metric = (
        random_ds(time_coords=True).isel(time=slice(0, -1)).rename({"data": metricname})
    )
    ds_metric.attrs = attrs
    with pytest.warns(UserWarning) as warninfo:
        ddict_matched = match_metrics(
            {"a": ds},
            {"aa": ds_metric},
            [metricname],
            print_statistics=True,
            dim_length_conflict="align",
        )
    msg = "none.none.a.a.a.a.a.a.a:`metric` dimensions ['time:5'] do not match `ds` ['time:6']. Aligning the data on `inner`"
    assert warninfo[0].message.args[0] == msg

    xr.testing.assert_allclose(ddict_matched["a"].time, ds_metric.time)


def test_match_metrics_print_statistics(capsys, metricname):
    # give a dataset that has member_id as dim (indicator that it was aggregated).

    attrs = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }
    ds = random_ds()
    ds.attrs = attrs
    ds_metric = random_ds().rename({"data": metricname})
    ds_metric.attrs = attrs

    match_metrics({"a": ds}, {"aa": ds_metric}, [metricname], print_statistics=True)

    captured = capsys.readouterr()

    assert "Processed 1 datasets." in captured.out
    assert "Exact matches:" + str({metricname: 1}) in captured.out
    assert "Other matches:" + str({metricname: 0}) in captured.out
    assert "No match found:" + str({metricname: 0}) in captured.out


def test_match_metrics_match_variable_str_input():
    # give a dataset that has member_id as dim (indicator that it was aggregated).
    metricname = "area"
    attrs = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }
    ds = random_ds()
    ds.attrs = attrs
    ds_metric = random_ds().rename({"data": metricname})
    ds_metric.attrs = attrs

    ds_dict_parsed_list = match_metrics({"a": ds}, {"aa": ds_metric}, [metricname])
    ds_dict_parsed_str = match_metrics({"a": ds}, {"aa": ds_metric}, metricname)
    xr.testing.assert_equal(ds_dict_parsed_str["a"], ds_dict_parsed_list["a"])


@pytest.mark.parametrize("combine_func_kwargs", [{}, {"compat": "override"}])
def test_combine_datasets_merge(combine_func_kwargs):
    attrs_a = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }

    attrs_b = {
        "source_id": "b",
        "grid_label": "b",
        "experiment_id": "b",
        "table_id": "b",
        "variant_label": "b",
        "version": "b",
    }

    # Create some datasets with a/b attrs
    ds_a_temp = random_ds(attrs=attrs_a).rename({"data": "temp"})
    ds_a_salt = random_ds(attrs=attrs_a).rename({"data": "salt"})

    ds_b_temp = random_ds(attrs=attrs_b).rename({"data": "temp"})
    ds_b_salt = random_ds(attrs=attrs_b).rename({"data": "salt"})
    ds_b_other = random_ds(attrs=attrs_b).rename({"data": "other"})

    ds_dict = {
        "".join(random.choices(string.ascii_letters, k=4)): ds
        for ds in [ds_a_salt, ds_a_temp, ds_b_salt, ds_b_temp, ds_b_other]
    }

    # Group together the expected 'matches'
    expected = {
        "a.a.a.a.a": [ds_a_salt, ds_a_temp],
        "b.b.b.b.b": [ds_b_other, ds_b_temp, ds_b_salt],
    }

    result = combine_datasets(
        ds_dict,
        xr.merge,
        combine_func_kwargs=combine_func_kwargs,
    )
    for k in expected.keys():
        assert k in list(result.keys())
        xr.testing.assert_equal(result[k], xr.merge(expected[k], **combine_func_kwargs))


def test_merge_variables():
    attrs_a = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }

    attrs_b = {
        "source_id": "b",
        "grid_label": "b",
        "experiment_id": "b",
        "table_id": "b",
        "variant_label": "b",
        "version": "b",
    }

    # Create some datasets with a/b attrs
    ds_a_temp = random_ds(attrs=attrs_a).rename({"data": "temp"})
    ds_a_salt = random_ds(attrs=attrs_a).rename({"data": "salt"})

    ds_b_temp = random_ds(attrs=attrs_b).rename({"data": "temp"})
    ds_b_salt = random_ds(attrs=attrs_b).rename({"data": "salt"})
    ds_b_other = random_ds(attrs=attrs_b).rename({"data": "other"})

    ds_dict = {
        "".join(random.choices(string.ascii_letters, k=4)): ds
        for ds in [ds_a_salt, ds_a_temp, ds_b_salt, ds_b_temp, ds_b_other]
    }

    # Group together the expected 'matches'
    expected = {
        "a.a.a.a.a": [ds_a_salt, ds_a_temp],
        "b.b.b.b.b": [ds_b_other, ds_b_temp, ds_b_salt],
    }

    result = merge_variables(ds_dict)

    for k in expected.keys():
        assert k in list(result.keys())
        xr.testing.assert_equal(
            result[k],
            xr.merge(
                expected[k],
                **{
                    "compat": "override",
                    "join": "exact",
                    "combine_attrs": "drop_conflicts",
                },
            ),
        )


@pytest.mark.parametrize("concat_kwargs", [{}, {"compat": "override"}])
def test_concat_members(concat_kwargs):
    concat_kwargs = {}

    attrs_a = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }

    attrs_b = {k: v for k, v in attrs_a.items()}
    attrs_b["variant_label"] = "b"

    attrs_c = {k: v for k, v in attrs_b.items()}
    attrs_c["source_id"] = "c"

    # Create some datasets with a/b attrs
    ds_a_temp = random_ds(attrs=attrs_a).rename({"data": "temp"})
    ds_b_temp = random_ds(attrs=attrs_b).rename({"data": "temp"})

    ds_c_other = random_ds(attrs=attrs_c).rename({"data": "other"})

    ds_dict = {
        "".join(random.choices(string.ascii_letters, k=4)): ds
        for ds in [ds_a_temp, ds_b_temp, ds_c_other]
    }

    # Group together the expected 'matches'
    # promote the member_id like in concat_members
    expected = {
        "a.a.a.a": [
            _construct_and_promote_member_id(ds_a_temp),
            _construct_and_promote_member_id(ds_b_temp),
        ],
        "c.a.a.a": [_construct_and_promote_member_id(ds_c_other)],
    }

    result = concat_members(
        ds_dict,
        concat_kwargs=concat_kwargs,
    )
    for k in expected.keys():
        assert k in list(result.keys())
        xr.testing.assert_equal(
            result[k],
            xr.concat(expected[k], "member_id", **concat_kwargs),
        )
    # assert that member_id is a proper coordinate
    assert "member_id" in result["a.a.a.a"].coords
    for member in ["a", "b"]:
        assert member in result["a.a.a.a"].member_id


def test_concat_members_existing_member_dim():
    attrs_a = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }

    attrs_b = {k: v for k, v in attrs_a.items()}
    attrs_b["variant_label"] = "b"

    # Create some datasets with a/b attrs
    ds_a = random_ds(attrs=attrs_a).rename({"data": "temp"})
    ds_b = random_ds(attrs=attrs_b).rename({"data": "temp"})

    ds_a_promoted = ds_a.expand_dims({"member_id": [ds_a.attrs["variant_label"]]})
    ds_b_promoted = ds_b.expand_dims({"member_id": [ds_b.attrs["variant_label"]]})

    # testing mixed case
    ds_dict = {"some": ds_a_promoted, "thing": ds_b}

    # promote the member_id like in concat_members
    expected = xr.concat([ds_a_promoted, ds_b_promoted], "member_id")

    result = concat_members(
        ds_dict,
    )

    xr.testing.assert_equal(
        result["a.a.a.a"],
        expected,
    )


def test_concat_members_existing_member_dim_different_warning():
    attrs_a = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }

    attrs_b = {k: v for k, v in attrs_a.items()}
    attrs_b["variant_label"] = "b"

    # Create some datasets with a/b attrs
    ds_a = random_ds(attrs=attrs_a).rename({"data": "temp"})
    ds_b = random_ds(attrs=attrs_b).rename({"data": "temp"})

    ds_a_promoted_wrong = ds_a.expand_dims({"member_id": ["something"]})

    # testing mixed case
    ds_dict = {"some": ds_a_promoted_wrong, "thing": ds_b}
    msg = "but this is different from the reconstructed value"
    # TODO: Had trouble here when putting in the actual values I expected.
    # Probably some regex shit. This should be enough for now
    with pytest.warns(UserWarning, match=msg):
        concat_members(
            ds_dict,
        )


def test_concat_members_reconstruct_from_sub_experiment_id():
    attrs_a = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }

    attrs_b = {k: v for k, v in attrs_a.items()}
    attrs_b["variant_label"] = "b"
    attrs_b["sub_experiment_id"] = "sub_something"

    # Create some datasets with a/b attrs
    ds_a = random_ds(attrs=attrs_a).rename({"data": "temp"})
    ds_b = random_ds(attrs=attrs_b).rename({"data": "temp"})

    ds_a_promoted = ds_a.expand_dims({"member_id": ["a"]})
    ds_b_promoted = ds_b.expand_dims({"member_id": ["sub_something-b"]})

    # testing mixed case
    ds_dict = {"some": ds_a, "thing": ds_b}

    # promote the member_id like in concat_members
    expected = xr.concat([ds_a_promoted, ds_b_promoted], "member_id")

    result = concat_members(
        ds_dict,
    )

    xr.testing.assert_equal(
        result["a.a.a.a"],
        expected,
    )


@pytest.mark.parametrize("concat_kwargs", [{}, {"compat": "override"}])
def test_concat_experiments(concat_kwargs):
    concat_kwargs = {}

    attrs_a = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
    }

    attrs_b = {k: v for k, v in attrs_a.items()}
    attrs_b["experiment_id"] = "b"

    attrs_c = {k: v for k, v in attrs_b.items()}
    attrs_c["source_id"] = "c"

    # Create some datasets with a/b attrs
    ds_a_temp = random_ds(attrs=attrs_a).rename({"data": "temp"})
    ds_a_temp = ds_a_temp.assign_coords(
        time=xr.cftime_range("2000", periods=len(ds_a_temp.time))
    )

    ds_b_temp = random_ds(attrs=attrs_b).rename({"data": "temp"})
    ds_b_temp = ds_b_temp.assign_coords(
        time=xr.cftime_range("1850", periods=len(ds_b_temp.time))
    )

    ds_c_other = random_ds(attrs=attrs_c).rename({"data": "other"})
    ds_c_other = ds_c_other.assign_coords(
        time=xr.cftime_range("2000", periods=len(ds_c_other.time))
    )

    ds_dict = {
        "".join(random.choices(string.ascii_letters, k=4)): ds
        for ds in [ds_a_temp, ds_b_temp, ds_c_other]
    }

    # Group together the expected 'matches'
    expected = {
        "a.a.a.a": [ds_b_temp, ds_a_temp],  # these should be sorted in this order!
        "c.a.a.a": [ds_c_other],
    }

    result = concat_experiments(
        ds_dict,
        concat_kwargs=concat_kwargs,
    )
    for k in expected.keys():
        assert k in list(result.keys())
        xr.testing.assert_equal(
            result[k], xr.concat(expected[k], "time", **concat_kwargs)
        )

    # shuffle the dict entries (not sure that is actually then processed in that order...)
    ds_dict = {
        "".join(random.choices(string.ascii_letters, k=4)): ds
        for ds in [ds_b_temp, ds_c_other, ds_a_temp]
    }

    # Group together the expected 'matches'
    expected = {
        "a.a.a.a": [ds_b_temp, ds_a_temp],  # these should be sorted in this order!
        "c.a.a.a": [ds_c_other],
    }

    result = concat_experiments(
        ds_dict,
        concat_kwargs=concat_kwargs,
    )
    for k in expected.keys():
        assert k in list(result.keys())
        xr.testing.assert_equal(
            result[k], xr.concat(expected[k], "time", **concat_kwargs)
        )


def test_pick_first_member():

    attrs_a = {
        "source_id": "a",
        "grid_label": "a",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a1b2",  # TODO: I might have to reevaluate this for some of the damip experiments (the only ones where variant_label!=member_id)
        "version": "a",
        "variable_id": "a",
    }

    attrs_b = {k: v for k, v in attrs_a.items()}
    attrs_b["variant_label"] = "a1b1"

    attrs_c = {k: v for k, v in attrs_b.items()}
    attrs_c["source_id"] = "b"
    attrs_c["variant_label"] = "a1b1"

    attrs_d = {k: v for k, v in attrs_b.items()}
    attrs_d["source_id"] = "b"
    attrs_d["variant_label"] = "a2b1"

    # Create some datasets with a/b attrs
    ds_a = random_ds(attrs=attrs_a).rename({"data": "temp"})
    ds_b = random_ds(attrs=attrs_b).rename({"data": "temp"})
    ds_c = random_ds(attrs=attrs_c).rename({"data": "temp"})
    ds_d = random_ds(attrs=attrs_d).rename({"data": "temp"})

    ds_dict = {
        "".join(random.choices(string.ascii_letters, k=4)): ds
        for ds in [ds_a, ds_b, ds_c, ds_d]
    }

    # Group together the expected 'matches'
    expected = {
        "a.a.a.a.a": ds_b,
        "b.a.a.a.a": ds_c,
    }

    result = pick_first_member(
        ds_dict,
    )
    print(result)
    print(result.keys())
    for k in expected.keys():
        assert k in list(result.keys())
        xr.testing.assert_equal(result[k], expected[k])


@pytest.mark.parametrize("verbose", [True, False])
def test_interpolate_grid_label(verbose):
    # build three datasets. Basically both datasets are available on grid_label hr (high res), but only one on grid_label lr (low res)
    ds_lr_vara = xr.DataArray(
        np.random.rand(4, 5, 6),
        dims=["x", "y", "lev"],
        coords={
            "lon": ("x", np.linspace(0, 360, 4)),
            "lat": ("y", np.linspace(-90, 90, 5)),
        },
    ).to_dataset(name="vara")
    ds_lr_vara.attrs = {
        "source_id": "a",
        "grid_label": "lr",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
        "variable_id": "vara",
    }

    ds_hr_vara = xr.DataArray(
        np.random.rand(10, 50, 6),
        dims=["x", "y", "lev"],
        coords={
            "lon": ("x", np.linspace(0, 360, 10)),
            "lat": (["y"], np.linspace(-90, 90, 50)),
        },
    ).to_dataset(name="vara")

    ds_hr_vara.attrs = {
        "source_id": "a",
        "grid_label": "hr",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
        "variable_id": "vara",
    }

    ds_hr_varb = xr.DataArray(
        np.random.rand(10, 50, 6),
        dims=["x", "y", "lev"],
        coords={
            "lon": ("x", np.linspace(0, 360, 10)),
            "lat": (["y"], np.linspace(-90, 90, 50)),
        },
    ).to_dataset(name="varb")

    ds_hr_varb.attrs = {
        "source_id": "a",
        "grid_label": "hr",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
        "variable_id": "varb",
    }

    # put in a dataset that has consistent grid_label

    ds_simple_vara = xr.DataArray(
        np.random.rand(10, 50, 6),
        dims=["x", "y", "lev"],
        coords={
            "lon": ("x", np.linspace(0, 360, 10)),
            "lat": (["y"], np.linspace(-90, 90, 50)),
        },
    ).to_dataset(name="vara")

    ds_simple_vara.attrs = {
        "source_id": "b",
        "grid_label": "hr",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
        "variable_id": "vara",
    }

    ds_simple_varb = xr.DataArray(
        np.random.rand(10, 50, 6),
        dims=["x", "y", "lev"],
        coords={
            "lon": ("x", np.linspace(0, 360, 10)),
            "lat": (["y"], np.linspace(-90, 90, 50)),
        },
    ).to_dataset(name="varb")

    ds_simple_varb.attrs = {
        "source_id": "b",
        "grid_label": "hr",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
        "variable_id": "varb",
    }

    # put in another rando dataset

    ds_rando = xr.DataArray(
        np.random.rand(14, 55, 6),
        dims=["x", "y", "lev"],
        coords={
            "lon": ("x", np.linspace(0, 360, 14)),
            "lat": (["y"], np.linspace(-90, 90, 55)),
        },
    ).to_dataset(name="varb")

    ds_rando.attrs = {
        "source_id": "b",
        "grid_label": "blubb",
        "experiment_id": "a",
        "table_id": "a",
        "variant_label": "a",
        "version": "a",
        "variable_id": "varb",
    }

    ddict = {
        "a_lr": ds_lr_vara,
        "a_hr": ds_hr_vara,
        "b_hr": ds_hr_varb,
        "a_simple": ds_simple_vara,
        "b_simple": ds_simple_varb,
        "rando": ds_rando,
    }

    expected_simple = xr.merge([ds_simple_vara, ds_simple_varb])

    # Prefer the high res version (no interpolation needed)
    combined_ddict = interpolate_grid_label(
        ddict, target_grid_label="hr", verbose=verbose
    )

    expected = xr.merge([ds_hr_vara, ds_hr_varb], combine_attrs="drop_conflicts")

    xr.testing.assert_allclose(combined_ddict["a.a.a.a"], expected)
    xr.testing.assert_allclose(combined_ddict["b.a.a.a"], expected_simple)

    # now the other way around (interpolation needed for variable whatever_else)
    combined_ddict = interpolate_grid_label(ddict, target_grid_label="lr")

    regridder = xesmf.Regridder(
        ds_hr_varb, ds_lr_vara, "bilinear", periodic=True, ignore_degenerate=True
    )

    regridded = regridder(ds_hr_varb)
    expected = xr.merge([ds_lr_vara, regridded])

    xr.testing.assert_allclose(combined_ddict["a.a.a.a"], expected)
    xr.testing.assert_allclose(combined_ddict["b.a.a.a"], expected_simple)

    assert combined_ddict["a.a.a.a"].varb.attrs["xmip_regrid_method"] == "bilinear"


def test_nested_operations():
    # set up a few datasets to combine
    ds_1 = xr.Dataset(
        {"a": ("x", np.array([1]))},
        attrs={
            "source_id": "a",
            "variant_label": "a",
            "variable_id": "a",
            "experiment_id": "a",
        },
    )
    ds_2 = xr.Dataset(
        {"a": ("x", np.array([2]))},
        attrs={
            "source_id": "a",
            "variant_label": "b",
            "variable_id": "a",
            "experiment_id": "a",
        },
    )
    ds_3 = xr.Dataset(
        {"b": ("x", np.array([3]))},
        attrs={
            "source_id": "a",
            "variant_label": "a",
            "variable_id": "b",
            "experiment_id": "a",
        },
    )
    ds_4 = xr.Dataset(
        {"b": ("x", np.array([4]))},
        attrs={
            "source_id": "a",
            "variant_label": "b",
            "variable_id": "b",
            "experiment_id": "a",
        },
    )
    ddict = {"ds1": ds_1, "ds2": ds_2, "ds3": ds_3, "ds4": ds_4}

    ddict = {k: _construct_and_promote_member_id(ds) for k, ds in ddict.items()}

    ds_expected = xr.Dataset(
        {
            "a": xr.DataArray(
                [1, 2], dims=["member_id"], coords={"member_id": ["a", "b"]}
            ).expand_dims("x"),
            "b": xr.DataArray(
                [3, 4], dims=["member_id"], coords={"member_id": ["a", "b"]}
            ).expand_dims("x"),
        },
        attrs={"source_id": "a", "experiment_id": "a"},
    )

    _, ds_combined = merge_variables(concat_members(ddict)).popitem()
    xr.testing.assert_allclose(ds_expected.squeeze(), ds_combined.squeeze())

    with pytest.warns(UserWarning) as warninfo:
        _, ds_combined = concat_members(merge_variables(ddict)).popitem()
    xr.testing.assert_allclose(ds_expected.squeeze(), ds_combined.squeeze())

    msg = (
        "Match attributes ['grid_label', 'table_id'] not found in any of the datasets."
    )
    assert msg in warninfo[0].message.args[0]



try:
    import intake
except ImportError:
    intake = None


def google_cmip_col(catalog="main"):
    """A tiny utility function to point to the 'official' pangeo cmip6 cloud files."""
    if intake is None:
        raise ImportError(
            "This functionality requires intake-esm. Install with `conda install -c conda-forge intake-esm"
        )
    if catalog == "main":
        return intake.open_esm_datastore(
            "https://storage.googleapis.com/cmip6/pangeo-cmip6.json"
        )
    # this doesnt work anymore, but ill leave it here as an example for the future
    #     elif catalog == "testing":
    #         return intake.open_esm_datastore(
    #             "https://storage.googleapis.com/cmip6/pangeo-cmip6-testing.json"
    #         )
    else:
        raise ValueError("Catalog not recognized. Should be `main` or `testing`")


def model_id_match(match_list, id_tuple):
    """Matches `id_tuple` to the list of tuples `exception_list`, which can contain
    wildcards (match any entry) and lists (match any entry that is in the list).

    Parameters
    ----------
    match_list : list
        list of tuples with id strings corresponding to e.g. `source_id`, `grid_label`...
    id_tuple : tuple
        single tuple with id strings.
    """
    # Check the size of tuples
    if any([len(t) != len(id_tuple) for t in match_list]):
        raise ValueError(
            "Each tuple in `match_list` must have the same number of elements as `match_id`"
        )

    match_list_checked = []
    for ml in match_list:
        ml_processed = []
        for i in range(len(ml)):
            match_element = ml[i]
            if isinstance(match_element, str) and match_element != "*":
                match_element = [match_element]
            if id_tuple[i] in match_element or match_element == "*":
                ml_processed.append(True)
            else:
                ml_processed.append(False)
        match_list_checked.append(all(ml_processed))
    return any(match_list_checked)


def _key_from_attrs(ds, attrs, sep="."):
    return sep.join([ds.attrs[i] if i in ds.attrs.keys() else "none" for i in attrs])


def cmip6_dataset_id(
    ds,
    sep=".",
    id_attrs=[
        "activity_id",
        "institution_id",
        "source_id",
        "experiment_id",
        "variant_label",
        "table_id",
        "grid_label",
        "version",
        "variable_id",
    ],
):
    """Creates a unique string id for e.g. saving files to disk from CMIP6 output

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset
    sep : str, optional
        String/Symbol to seperate fields in resulting string, by default "."

    Returns
    -------
    str
        Concatenated
    """
    return _key_from_attrs(ds, id_attrs, sep=sep)


def _maybe_make_list(item):
    "utility function to make sure output is a list"
    if isinstance(item, str):
        return [item]
    elif isinstance(item, list):
        return item
    else:
        return list(item)



import numpy as np
import pytest
import xarray as xr

from xmip.regionmask import _default_merge_dict, merged_mask


regionmask = pytest.importorskip(
    "regionmask", minversion="0.5.0+dev"
)  # All tests get skipped if the version of regionmask is not > 0.5.0


@pytest.mark.parametrize("verbose", [True, False])
def test_merge_mask(verbose):
    x = np.linspace(0, 360, 720)
    y = np.linspace(-90, 90, 360)
    data = np.random.rand(len(x), len(y))
    ds = xr.DataArray(data, coords=[("x", x), ("y", y)]).to_dataset(name="data")
    ds["lon"] = ds["x"] * xr.ones_like(ds["y"])
    ds["lat"] = xr.ones_like(ds["x"]) * ds["y"]

    basins = regionmask.defined_regions.natural_earth_v4_1_0.ocean_basins_50

    mask = merged_mask(basins, ds, verbose=verbose)

    # check if number of regions is correct
    mask_regions = np.unique(mask.data.flat)
    mask_regions = mask_regions[~np.isnan(mask_regions)]

    assert len(mask_regions) == len(_default_merge_dict().keys())

    # now a brief range check to make sure the pacific is stamped out correctly
    pac = ds.where(
        np.logical_or(np.logical_or(mask == 2, mask == 3), mask == 4), drop=True
    )
    assert pac.lon.min() > 95.0
    assert pac.lon.max() < 295.0

    # I shoud add a test for -180-180

    # How to use cloud data.



try:
    from importlib.metadata import (  # only works for python 3.8 and upwards
        PackageNotFoundError,
        version,
    )
except:
    from importlib_metadata import (  # works for python <3.8
        PackageNotFoundError,
        version,
    )

try:
    __version__ = version("xmip")
except PackageNotFoundError:
    # package is not installed
    __version__ = "unknown"
    pass
